# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateHoxSEM(Template):
    'RP Note: Not too sure what this does, but original description was the same as for FactorAnalysis template. IMPORTANT: if using the e-STAT engine, you need to opt for "No optimisation steps" (see Settings via the main page).'

    __version__ = '1.0.0'

    tags = [ 'Model' , '1-Level']
    engines = ['eStat']

    inputs = '''
y = DataMatrix('Responses: ')

expl = Boolean('Are there any global fixed effects?: ')
if expl:
    x = DataMatrix('Explanatory variables: ', allow_cat = True)
NumFact = Integer('Number of factors: ')
CorrFact = Boolean('Are factors correlated?: ')
for j in range(0, int(NumFact)):
    context['fact' + str(j+1)] = ParamVector()
    for i in range(0, len(y)):
        context['constr_load' + str(j+1) + '_' + str(i+1)] = Boolean('Is loading ' + str(i+1) + ' for factor ' + str(j+1) + ' constrained? ')
        if not context['constr_load' + str(j+1) + '_' + str(i+1)]:
            context['load' + str(j+1) + '_' + str(i+1)] = ParamScalar(monitor=False)
        else:
            context['load' + str(j+1) + '_' + str(i+1)] = Text('value:')

    context['estmean' + str(j+1)] = Boolean('Estimate factor mean for factor ' + str(j + 1) + '?:')
    if context['estmean' + str(j+1)]:
        context['nu' + str(j+1)] = ParamScalar()

    if not CorrFact:
        context['constr_fact' + str(j+1)] = Boolean('Is factor ' + str(j+1) + ' variance constrained? ')
        if not context['constr_fact' + str(j+1)]:
            context['tau_fact' + str(j+1)] = ParamScalar()
            context['sigma2_fact' + str(j+1)] = ParamScalar(modelled = False)
        else:
            context['tau_fact' + str(j+1)] = Text('Value: ')

for i in range(0, len(y)):  
    context['miss'+y[i]] = ParamVector(monitor=False)

if expl:
    for i in range(0, len(y)):
        context['beta' + '_' + str(i+1)] = ParamVector(parents=[x], as_scalar=True)

if CorrFact:
    omega_fact = ParamMatrix(customstep=True)
    omega_fact.size = int(NumFact)
    d_fact = ParamMatrix(customcode=True)
    d_fact.size = int(NumFact)
    priors = Text('Priors for factor variance matrix: ', ['Uniform', 'Wishart'])
    if priors.name == 'Wishart':
        R = Text('R matrix: ')
        v = Text('Degrees of Freedom:')

equalvar = Boolean('Do you wish to constrain responses to have the same variance?: ')
if not equalvar:
    for i in range(0, len(y)):  
        context['tau_u2_' + str(i+1)] = ParamScalar()
        context['sigma2_u2_' + str(i+1)] = ParamScalar(modelled = False)
else:
    tau_u2 = ParamScalar()
    sigma2_u2 = ParamScalar(modelled = False)

deviance = ParamScalar(modelled = False)
'''      

    model = '''
<% numfacts = int(NumFact) %>
model {
    for (i in 1:length(${y[0]})) {
        % for i in range(0, len(y)):
% if equalvar:
        miss${y[i]}[i] ~ dnorm(mu${i+1}[i], tau_u2)
% else:
        miss${y[i]}[i] ~ dnorm(mu${i+1}[i], tau_u2_${i+1})
% endif
% if expl:
        mu${i+1}[i] <- ${mmult(x, 'beta' + str(i+1) + '_', 'i')} \\
        % for j in range(0, numfacts):
        + load${j+1}_${i+1} * fact${j+1}[i] 
        % endfor
% else:
        mu${i+1}[i] <- load${1}_${i+1} * fact${1}[i] \\
        % for j in range(1, numfacts):
        + load${j+1}_${i+1} * fact${j+1}[i] 
        % endfor
% endif
        dummy${y[i]}[i] ~ ddummy(miss${y[i]}[i])
        % endfor
    }
    % for i in range(0, len(y)):
    % for j in range(0, numfacts):
    % if not context['constr_load' + str(j+1) + '_' + str(i+1)]:
    load${j+1}_${i+1} ~ dflat()
    % else:
    load${j+1}_${i+1} <- ${context['load' + str(j+1) + '_' + str(i+1)]}
    % endif
    % endfor
    % endfor
    
    for (i in 1:length(${y[0]})) {
% if CorrFact:
<% num = int(NumFact) %> \\
          dummy[i] ~ dnormal${num}a( \\
           % for j in range(0, num):
fact${j+1}[i], \\
           % endfor
           % for j in range(0, num):
% if context['estmean' + str(j+1)]:
nu${j + 1}, \\
% else:
0, \\
%endif
           % endfor
           % for j in range(0,(((num+1)*num)/2) - 1):
d_fact[${j}], \\
           % endfor
d_fact[${(((num+1)*num)/2) - 1}])
           % for j in range(0, num):
           fact${j+1}[i] ~ dflat()
           % endfor

% else:
        % for j in range(0, numfacts):
% if context['estmean' + str(j+1)]:
        fact${j+1}[i] ~ dnorm(nu${j + 1}, tau_fact${j+1})
% else:
        fact${j+1}[i] ~ dnorm(0, tau_fact${j+1})
% endif
        % endfor
% endif
    }
    
    # Priors
% if expl:
    % for i in range(0, len(y)):
    % for j in range(0, len(x)):
    beta_${i+1}_${j} ~ dflat()
    % endfor
    % endfor
% endif

% for j in range(0, numfacts):
% if context['estmean' + str(j+1)]:
    nu${j + 1} ~ dflat()
% endif
% endfor

% if equalvar:
    tau_u2 ~ dgamma(0.001000, 0.001000)
    sigma2_u2 <- 1 / tau_u2    
% else:
    % for i in range(0, len(y)):
    tau_u2_${i+1} ~ dgamma(0.001000, 0.001000)
    sigma2_u2_${i+1} <- 1 / tau_u2_${i+1} 
    % endfor
% endif
%if not CorrFact:
    % for j in range(0, numfacts):
    % if not context['constr_fact' + str(j+1)]:
    tau_fact${j+1} ~ dgamma(0.001000, 0.001000)
    sigma2_fact${j+1} <- 1 / tau_fact${j+1}
    % else:
    tau_fact${j+1} <- ${context['tau_fact' + str(j+1)]}
    % endif
    % endfor
% endif
}
'''
    
    preccode = '''
            {
% if CorrFact:
<% n = int(NumFact) %>\\
                // Note currently using a uniform prior for variance matrix
                RectMatrix sb(${n},${n});
                for(int i = 0; i < length(fact1); i++) {
% for j in range(0, n):
% for k in range(0, n):
% if context['estmean' + str(j+1)] and context['estmean' + str(k+1)]:
                    sb(${j}, ${k}) += double(fact${j + 1}[i] - nu${j + 1}) * double(fact${k + 1}[i] - nu${k + 1});
% endif

% if context['estmean' + str(j+1)] and not context['estmean' + str(k+1)]:
                    sb(${j}, ${k}) += double(fact${j + 1}[i] - nu${j + 1}) * double(fact${k + 1}[i]);
% endif

% if not context['estmean' + str(j+1)] and context['estmean' + str(k+1)]:
                    sb(${j}, ${k}) += double(fact${j + 1}[i]) * double(fact${k + 1}[i] - nu${k + 1});
% endif

% if not context['estmean' + str(j+1)] and not context['estmean' + str(k+1)]:
                    sb(${j}, ${k}) += double(fact${j + 1}[i]) * double(fact${k + 1}[i]);
% endif


% endfor
% endfor
                }
% if priors.name == 'Uniform':
                int vw = length(fact1)-${n + 1};
                if (runstate == 0) {
% for j in range(0, n):
                    sb(${j}, ${j}) += 0.0001;
% endfor
                }
% endif
% if priors.name == 'Wishart':
                int vw = length(fact1) + ${v.name};
% endif

% if priors.name == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for j in range(0, n):
    for k in range(0, j + 1):
        Rmat[j, k] = float(R.name[count]) 
        Rmat[k, j] = Rmat[j, k]
        count += 1
%>

% for j in range(0, n):
% for k in range(0, n):
                sb(${j}, ${k}) += ${str(Rmat[j, k] * float(v.name))};
% endfor
% endfor
% endif

                matrix_sym_invinplace(sb);
                bool fail = false;
                mat_d_fact = dwishart(vw, sb, fail); 
                mat_omega_fact = matrix_sym_inverse(mat_d_fact, fail);
%endif
            }
         '''

    preparedata = '''
mydata = data['datafile']
for i in range(0, int(NumFact)):
    context['fact' + str(i+1)].ncols = -1*len(mydata.variables[y[0]]['data'])
            
for i in range(0, len(y)):
    context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
'''
