# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateNLevXCFact(Template):
    'Fits a factor analysis model with one or more higher classifications (nested or crossed); uses random intercepts only. IMPORTANT: if using the eStat engine, you need to opt for "No optimisation steps" (see Settings via the main page).'

    __version__ = '1.0.0'

    tags = [ 'Model', 'N-Level', 'Normal', 'Factor analysis' ]
    engines = ['eStat', 'MLwiN_MCMC']

    inputs = '''
D = Text(value = 'MVNormal')
NumLevs = Integer('Number of Classifications: ')
for i in range(0, int(NumLevs)):
    selstr = 'Classification ' + str(i + 1) + ': '
    context['C' + str(i + 1)] = IDVector(selstr)
y = DataMatrix('Responses: ')
x = DataMatrix('Explanatory variables: ', allow_cat = True)
NumFact = Integer('Number of factors: ')
for j in range(0, int(NumFact)):
    context['fact_level' + str(j)] = Integer('Classification number for factor ' + str(j + 1) + ' (0 for observation level) : ')
    context['fact' + str(j)] = ParamVector(monitor=False)
    for i in range(0, len(y)):
        context['constr_load' + str(j) + '_' + str(i)] = Boolean('Is loading ' + str(i+1) + ' for factor ' + str(j+1) + ' constrained? ')
        if not context['constr_load' + str(j) + '_' + str(i)]:
            context['load' + str(j) + '_' + str(i)] = ParamScalar()
        else:
            context['load' + str(j) + '_' + str(i)] = Text('value:')
          
    context['constr_fact' + str(j)] = Boolean('Is factor ' + str(j+1) + ' variance constrained? ')
    if not context['constr_fact' + str(j)]:
        context['tau_fact' + str(j)] = ParamScalar()
        context['sigma2_fact' + str(j)] = ParamScalar(modelled = False)
    else:
        context['tau_fact' + str(j)] = Text('Value: ')

storeresid = Boolean('Store residuals?')

for i in range(0, len(y)):
    context['miss'+y[i]] = ParamVector(monitor=False)
    context['tau_' + str(i)] = ParamScalar()
    context['sigma2_' + str(i)] = ParamScalar(modelled = False)
    context['beta' + '_' + str(i)] = ParamVector(parents=[x], as_scalar=True)
    for j in range(0, int(NumLevs)):
        if storeresid:
            context['u' + str(j + 1) + '_' + str(i)] = ParamVector(parents=[context['C' + str(j + 1)]], as_scalar=False)
        else:
            context['u' + str(j + 1) + '_' + str(i)] = ParamVector(parents=[context['C' + str(j + 1)]], as_scalar=False, monitor=False)
        context['tau_u' + str(j + 1) + '_' + str(i)] = ParamScalar()   
        context['sigma2_u' + str(j + 1) + '_' + str(i)] = ParamScalar(modelled = False)   

deviance = ParamScalar(modelled=False, customstep=True)
'''

    model = '''
<% numfacts = int(NumFact) %> \\
<% numlevs = int(NumLevs) %> \\
model {
        for (i in 1:length(${y[0]})) {
            % for n in range(0, len(y)):
            miss${y[n]}[i] ~
            dnorm(mu${n}[i], tau_${n})
            mu${n}[i] <-
            ${mmult(x, 'beta' + str(n) + '_', 'i')}
            % for j in range(0, numfacts):
            <% factclass = int(context['fact_level' + str(j)]) %>
            % if factclass > 0:
            + load${j}_${n} * fact${j}[${context['C' + str(factclass)]}[i]]
            % else:
            + load${j}_${n} * fact${j}[i]
            % endif
            % endfor
            % for i in range(0, numlevs):
            + u${i + 1}_${n}[${context['C' + str(i + 1)]}[i]]
            % endfor
            dummy${y[n]}[i] ~ ddummy(miss${y[n]}[i])
            % endfor
        }
    
    % for i in range(0, len(y)):
    % for j in range(0, numfacts):
    % if not context['constr_load' + str(j) + '_' + str(i)]:
    load${j}_${i} ~ dflat()
    % else:
    load${j}_${i} <- ${context['load' + str(j) + '_' + str(i)]}
    % endif
    % endfor
    % endfor
    
    % for j in range(0, numfacts):
    <% factclass = int(context['fact_level' + str(j)]) %>
    % if factclass > 0:
    for (i${factclass} in 1:length(fact${j})) {
        fact${j}[i${factclass}] ~ dnorm(0, tau_fact${j})
    }
    % else:
    for (i in 1:length(fact${j})) {
        fact${j}[i] ~ dnorm(0, tau_fact${j})
    }

    % endif
    % endfor
    
    % for i in range(0, numlevs):
    for (i${i + 1} in 1:length(u${i + 1}_0)) {
        % for n in range(0, len(y)):
        u${i + 1}_${n}[i${i + 1}] ~ dnorm(0, tau_u${i + 1}_${n})
        % endfor
    }
    % endfor
        
    # Priors
    
    % for i in range(0, len(y)):
    % for j in range(0, x.ncols()):
    beta_${i}_${j} ~ dflat()
    % endfor
    % endfor
    
    % for i in range(0, len(y)):
    tau_${i} ~ dgamma(0.001000, 0.001000)
    sigma2_${i} <- 1 / tau_${i}
    % for j in range(0, numlevs):
    tau_u${j + 1}_${i} ~ dgamma(0.001000, 0.001000)
    sigma2_u${j + 1}_${i} <- 1 / tau_u${j + 1}_${i}
    % endfor
    % endfor
    % for j in range(0, numfacts):
    % if not context['constr_fact' + str(j)]:
    tau_fact${j} ~ dgamma(0.001000, 0.001000)
    sigma2_fact${j} <- 1 / tau_fact${j}
    % else:
    tau_fact${j} <- ${context['tau_fact' + str(j)]}
    % endif
    % endfor
}        
'''

    latex = r'''
<% lenbeta = 0 %>
<% numlevs = int(NumLevs) %>
<% numfacts = int(NumFact) %>
\begin{aligned} 
% for i in range(0, len(y)):
\mbox{${y[i]}}_i & \sim \mbox{N} \left( \mu_{${i}i}, \sigma^2_{e${i}} \right) \\ 
% endfor
% for i in range(0, len(y)):
\mu_{${i}i} & = ${mmulttex2(x, r'\beta', 'i', lenbeta)} 
% for lev in range(0, numlevs):
 + u^{(${lev + 2})}_{${i},${context['C' + str(lev + 1)]}(i)} 
% endfor
% for j in range(0, numfacts):
+ 
% if not context['constr_load' + str(j) + '_' + str(i)]:
\lambda_{${j},${i}}
% else:
${context['load' + str(j) + '_' + str(i)]}
% endif
% if int(context['fact_level' + str(j)]) > 0:
\eta^{(${context['fact_level' + str(j)] + 1})}_{${j}${context['C' + str(context['fact_level' + str(j)])]}(i)}
% else:
\eta^{(${context['fact_level' + str(j)] + 1})}_{${j}i}
% endif
% endfor
<% lenbeta += len(x) %>
\\ 
% endfor
% for lev in range(0, numlevs):
% for i in range(0, len(y)):
u^{(${lev + 2})}_{${i},${context['C' + str(lev + 1)]}(i)} & \sim \mbox{N} \left(0, \sigma^2_{u${i}} \right) \\ 
% endfor
% endfor
% for i in range(0, len(y)):
\sigma^2_{e${i}} & \sim \Gamma(0.001000, 0.001000) \\ 
% endfor
% for j in range(0, numfacts):
% if int(context['fact_level' + str(j)]) > 0:
\eta^{(${context['fact_level' + str(j)] + 1})}_{${j}${context['C' + str(context['fact_level' + str(j)])]}(i)}
% else:
\eta^{(${context['fact_level' + str(j)] + 1})}_{${j}i}
% endif
& \sim \mbox{N} 
% if not context['constr_fact' + str(j)]:
\left(0, \sigma^2_{\eta${j}} \right) \\ 
% else:
\left(0, ${context['tau_fact' + str(j)]} \right) \\ 
% endif
% endfor
% for lev in range(0, numlevs):
% for i in range(0, len(y)):
\sigma^2_{u${i}} & \sim \Gamma(0.001000, 0.001000) \\ 
% endfor
% endfor
% for i in range(0, lenbeta):
\beta_{${i}} & \propto 1 \\ 
% endfor
% for j in range(0, numfacts):
% if not context['constr_fact' + str(j)]:
\sigma^2_{\eta${j}} & \sim \Gamma(0.001000, 0.001000) \\ 
% endif
% endfor
% for j in range(0, numfacts):
% for i in range(0, len(y)):
% if not context['constr_load' + str(j) + '_' + str(i)]:
\lambda_{${j},${i}} & \propto 1 \\ 
% endif
% endfor
% endfor
\end{aligned}
'''

    mlnscript = '''
<% N=datafile.nobs %>
% for i in range(len(y)):
MVAR 1 "${y[i]}"
% endfor
% for i in range(int(NumLevs)):
IDEN ${i + 3} "${context['C' + str(i + 1)]}"
% endfor
IDEN 2 "_id"
IDEN 1 "resp_indicator"

<%
import numpy
T=numpy.eye(len(y), dtype='int')
%>
% for j in range(0, len(T)):
RPAT ${' '.join([str(k) for k in T[j]])}
% for i in range(0,len(x)):
ADDT "${x[i]}"
% endfor
ADDT "_levres"
RPAT
% endfor

% for i in range(len(y)):
% for j in range(1, NumLevs + 2):
SETV ${j + 1} "_levres.${i + 1}"
% endfor
FPAR 0 "_levres.${i + 1}"
% endfor

<% TT='' %>
% for j in range(0, int(NumFact)):
<%TT=TT+ str(int(context['fact_level' + str(j)])+2) +' '%>
SMAT ${int(context['fact_level' + str(j)])+2} 0
<%
    for i in range(0, len(y)):
        if context['constr_load' + str(j) + '_' + str(i)]:
            TT=TT+' '+str(context['load' + str(j) + '_' + str(i)])+' '+str(int(bool(context['constr_load' + str(j) + '_' + str(i)])))+ ' '
        else:
            TT=TT+' '+str(0)+' '+str(int(bool(context['constr_load' + str(j) + '_' + str(i)])))+ ' '
    if context['constr_fact' + str(j)]:
        TT=TT+ str(context['tau_fact' + str(j)]) +' '+ str(int(bool(context['constr_fact' + str(j)])))+' '
    else:
        TT=TT+ str(1) +' '+ str(int(bool(context['constr_fact' + str(j)])))+' '

%>
% endfor
FACT ${len(y)} ${int(NumFact)} 0 ${TT}
XCLA 1
        '''
      
    preparedata = '''
mydata = data['datafile']
numfacts = NumFact

for j in range(0, numfacts):
    factclass = context['fact_level' + str(j)]
    if factclass > 0:
        context['fact' + str(j)].ncols = -1*len(numpy.unique(mydata.variables[context['C' + str(factclass)]]['data']))
    else:
        context['fact' + str(j)].ncols = -1*len(mydata.variables[y[0]]['data'])
for i in range(0, len(y)):
    context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
'''
